:- use_module(library(lists), [select/3]).
:- use_module('~/oktatas/Common/utils', [between/3]).
:- use_module(library(timeout)).


:- block sorted(-).
% sorted(L): L rendezett.
sorted([X|L]) :-
	sorted(L, X).

% sorted(L, X): [X|L] rendezett.
:- block sorted(-, ?).
sorted([Y|L], X) :-
	freeze(X, freeze(Y, X =< Y)), sorted(L, Y).
sorted([], _).

/* ugyanez, freeze nlkl: 
:- block sorted(-, ?), sorted(?, -).
% sorted(L, X): [X|L] rendezett.
sorted([Y|L], X) :- 
	sorted(X, Y, L).
sorted([], _).

:- block sorted(?, -, ?).
% sorted(X, Y, L): [X,Y|L] rendezett.
sorted(X, Y, L) :-
	X =< Y, sorted(L, Y).
*/

% L rendezettje S.
gen_test_sort(L, S) :-
	permutation(L, S),
	sorted(S).     % block deklarci felesleges

% L rendezettje S.
blocked_test_gen_sort(L, S) :-
	sorted(S),     % block deklarci fontos!
	permutation(L, S).

/* Idzet a lists knyvtrbl:  */

%   permutation(?List, ?Perm)
%   is true when Perm is a permutation of List.
permutation([], []).
permutation(List, [First|Perm]) :- 
	select(First, List, Rest),
	permutation(Rest, Perm).

% decr_seq(N, L): L = [N,...,2,1]
decr_seq(0, L) :- !, L = [].
decr_seq(N, [N|L]) :-
	N1 is N-1, decr_seq(N1, L).

% teszt-eljrsok:

test(Len) :-
	decr_seq(Len, L),
	TimeOut = 5000,
	timed_limited_call(gen_test_sort(L, _), TimeOut, T0),
	timed_limited_call(blocked_test_gen_sort(L, _), TimeOut, T1),
	timed_limited_call(sort(L, _), TimeOut, T2),
	format('~t~w~4|~t ~w~9+~t~w~9+~t~w~9+~n', [Len,T0,T1,T2]).

timed_limited_call(Goal, Limit, Time) :-
	statistics(runtime, _),
	time_out(call(Goal), Limit, Res),
	(   Res = success -> statistics(runtime, [_,Time])
	;   Time = timeout
	).

test_up_to(Max) :- between(1, Max, Len),
	test(Len),
	fail.
test_up_to(_).

test :-
	test_up_to(20).


/*
| ?- test.
   1        0        0        0
   2        0        0        0
   3        0        0        0
   4        0        0        0
   5       10        0        0
   6       30        0        0
   7      220       10        0
   8     1770       10        0
   9  timeout       30        0
  10  timeout       70        0
  11  timeout      160        0
  12  timeout      340        0
  13  timeout      720        0
  14  timeout     1540        0
  15  timeout     3250        0
  16  timeout  timeout        0
  17  timeout  timeout        0
  18  timeout  timeout        0
  19  timeout  timeout        0
  20  timeout  timeout        0

yes
*/
