%---------------------------------------------------------------------------

:- use_module(library(clpfd)).

time(Goal, T) :-
	statistics(runtime, _),
	Goal,
	statistics(runtime, [_,T]).

% | ?- X in (10..20)/\ (\{15}), Y in 6..sup, Z #= X+Y.
% X in(10..14)\/(16..20), Y in 6..sup,
% Z in 16..sup ? 
% yes
% | ?- X in 10..20, X #\= 15, Y in {2}, Z #= X*Y.
% Y = 2, X in(10..14)\/(16..20),
% Z in 20..40 ? 
% yes
% | ?-
%---------------------------------------------------------------------------

% SEND+MORE=MONEY.
send(SEND,MORE,MONEY):-
  List= [S,E,N,D,M,O,R,Y], domain(List, 0, 9),
  alldiff(List), S #\= 0, M#\= 0,
  SEND #=  1000*S+100*E+10*N+D,
  MORE #= 1000*M+100*O+10*R+E,
  MONEY #= 10000*M+1000*O+100*N+10*E+Y,
  SEND+MORE #= MONEY, my_labeling(List).

% my_labeling(Xs) lnyegben azonos a beptett
% labeling([], Xs) eljrssal.
my_labeling([]).
my_labeling([X|Xs]) :- indomain(X), my_labeling(Xs).

% Azonos a beptett all_different eljrssal
alldiff([]).
alldiff([X|Xs]) :- outof(X, Xs), alldiff(Xs).

outof(_, []).
outof(X, [Y|Ys]) :- X #\= Y, outof(X, Ys).

% | ?- send(S, M, MONEY).
% 
% M = 1085,
% S = 9567,
% MONEY = 10652 ? ;
% 
% no
% | ?- 

%---------------------------------------------------------------------------

:- use_module(library(lists), [nth/3]).

zebra(ZOwner, All):-
  All = [England,Spain,Japan,Norway,Italy,
         Green,Red,Yellow,Blue,White,
         Painter,Diplomat,Violinist,Doctor,Sculptor,
         Dog, Zebra, Fox, Snail, Horse,
         Juice,Water,Tea,Coffee,Milk],
  domain(All, 1, 5),
  alldiff([England,Spain,Japan,Norway,Italy]),
  alldiff([Green,Red,Yellow,Blue,White]),
  alldiff([Painter,Diplomat,Violinist,Doctor,Sculptor]),
  alldiff([Dog,Zebra,Fox,Snail,Horse]),
  alldiff([Juice,Water,Tea,Coffee,Milk]),
  England = Red,          Spain = Dog,
  Japan = Painter,        Italy = Tea,
  Norway = 1,             Green = Coffee,
  Green #= White+1,       Sculptor = Snail,
  Diplomat = Yellow,      Milk = 3,
  Violinist = Juice,      nextto(Norway, Blue),
  nextto(Fox, Doctor),    nextto(Horse, Diplomat),
  labeling([ff], All),
  nth(N, [England,Spain,Japan,Norway,Italy], Zebra),
  nth(N, [england,spain,japan,norway,italy], ZOwner).

nextto(A, B) :-  abs(A-B) #= 1.

% | ?- zebra(ZOwner, All), write(All).
% [3,4,5,1,2,5,3,1,2,4,5,1,4,2,3,4,5,1,3,2,4,1,2,5,3]
% All = [3,4,5,1,2,5,3,1,2,4|...],
% ZOwner = japan ? ;
% 
% no
% | ?- 

%---------------------------------------------------------------------------
% | ?- domain([A,B,C,D,E], 1, 3), 
%      A #> B, A #\= C, A #\= D, A #\= E,
%      B #\= C, B #\= D, C #\= E, D #< E.
%           
% A in 2..3, B in 1..2,
% C in 1..3, D in 1..2, E in 2..3 ? ;
% 
% no
% | ?- domain([A,B,C,D,E], 1, 3), 
%      A #> B, A #\= C, A #\= D, A #\= E,
%      B #\= C, B #\= D, C #\= E, D #< E,
%      labeling([], [A,B,C,D,E]).
%                
% A = 3, B = 2, C = 1, D = 1, E = 2 ? ;
% 
% no
% | ?- domain([A,B,C,D,E], 1, 3),
%      all_distinct([A,B,C]), all_distinct([A,C,E]),
%      A #> B, E #> D.
%      
% A = 3, B = 2, C = 1, D = 1, E = 2 ? ;
% 
% no
% | ?-

%---------------------------------------------------------------------------

:- use_module(library(clpfd)).

% A Qs lista N kirlyn biztonsgos elhelyezst
% mutatja egy N*N-es sakktbln. Ha a lista
% i. eleme j, akkor az i. kirlynt az i. sor j.
% oszlopba kell helyezni.
queens(N, Qs, Env):-
        length(Qs, N), domain(Qs, 1, N),
        safe(Qs, Env),
        labeling([ff],Qs).           % first-fail elv

% safe(Qs): A Qs lista a kirlynk biztonsgos
% elhelyezst rja le.
safe([], _Env).
safe([Q|Qs], Env):-
        no_attack(Qs, Q, 1, Env),
        safe(Qs, Env).

% no_attack(Qs, Q, I): A Qs lista ltal lert
% kirlynk egyike sem tmadja a Q oszlopban lev
% kiralynt, feltve hogy Q s Qs tvolsga I.
no_attack([], _, _, _).
no_attack([X|Xs], Y, I, Env):-
      no_threat(Env, X, Y, I),
      J is I+1, no_attack(Xs, Y, J, Env).

:- discontiguous no_threat/4.
% Az X s Y oszlopokban  I sortvolsgra lev
% kirlynk nem tmadjk egymst.
no_threat(no_threat1, X, Y, I) :-
        Y #\= X, Y #\= X-I, Y #\= X+I.

% | ?- queens(4, W, no_threat1).
% 
% W = [2,4,1,3] ? ;
% 
% W = [3,1,4,2] ? ;
% 
% no
% | ?- findall(W, queens(8, W, no_threat1), _All), length(_All, Len).
% 
% Len = 92 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

% | ?- X in {4,9}, Y in {2,3}, Z #= X-Y.    % intervallum-konz.:
%                 X in{4}\/{9}, Y in 2..3, Z in 1..7 ? 
% | ?- X in {4,9}, Y=2, Z #= X-Y.           % tartomny-konz.:
%                 Y = 2, X in{4}\/{9}, Z in{2}\/{7} ? 

%---------------------------------------------------------------------------

% Az L lista egy N hosszsg mgikus sorozat.
magikus(N, L) :-
        length(L, N),
        N1 is N-1, domain(L, 0, N1),
        elofordulasok(L, 0, L),
        labeling([], L).             % most felesleges

% elofordulasok([Ei, ...], i, Sor): Sor-ban i, i+1, ... 
% elfordulsainak szma Ei, Ei+1, ...
elofordulasok([], _, _).
elofordulasok([E|Ek], I, Sor) :-
        pontosan(I, Sor, E), J is I+1, elofordulasok(Ek, J, Sor).

% pontosan(I, L, E): Az I szm L-ben pontosan E-szer fordul el.
pontosan(I, L, 0) :- outof(I, L).
pontosan(I, [I|L], N) :- N #> 0, N1 #= N-1, pontosan(I, L, N1).
pontosan(I, [X|L], N) :- N #> 0, X #\= I,   pontosan(I, L, N).

% Hatkonysg-fokozs redundns (surrogate) constraint-ekkel.
magikus1(N, L) :-
        length(L, N), N1 is N-1, domain(L, 0, N1),
        osszege(L, S), call(S #= N),
        szorzatosszege(L, 0, SS), call(SS #= N),
        elofordulasok(L, 0, L).

osszege([], 0).
osszege([X|L], X+S) :- osszege(L, S).

szorzatosszege([], _, 0).
szorzatosszege([X|L], I, I*X+S) :- 
        J is I+1, szorzatosszege(L, J, S).


% | ?- time(magikus(9, L), T).
% 
% L = [5,2,1,0,0,1,0,0,0],
% T = 4560 ? 
% 
% yes
% | ?- time(magikus1(9, L), T).
% 
% L = [5,2,1,0,0,1,0,0,0],
% T = 50 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------
% | ?- X in 1..4, X #< Y #<=> B, X+Y #=9.
%         B = 1, X in 1..4, Y in 5..8 ? 
% | ?- X+Y #= Z #<=> B, X=1, Z=6, Y in 1..10, Y#\=5.
%         X = 1, Z = 6, Y in(1..4)\/(6..10), B in 0..1 ? 
%---------------------------------------------------------------------------

pontosan2(_, [], 0).
pontosan2(I, [X|L], N) :-
        X #= I #<=> B, N #= N1+B, pontosan2(I, L, N1).

%---------------------------------------------------------------------------

magikus2(N, L) :-
        length(L, N), N1 is N-1, domain(L, 0, N1),
        elofordulasok3(L, 0, L, Egyhat),
        sum(L, #=, N), 
        scalar_product(Egyhat, L, #=, N),
        labeling([], L).

elofordulasok3([], _, _, []).
elofordulasok3([E|Ek], I, Sor, [I|EH]) :-
        count(I, Sor, #=, E),
        J is I+1, elofordulasok3(Ek, J, Sor, EH).


% | ?- time(magikus1(30, L), T).
% 
% L = [26,2,1,0,0,0,0,0,0,0|...],
% T = 8470 ? 
% 
% yes
% | ?- time(magikus2(30, L), T).
% 
% L = [26,2,1,0,0,0,0,0,0,0|...],
% T = 370 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

% | ?- element(X, [0,1,2,3,4], Y), X in {2,5}. % Y #= X-1
%                         X in{2}\/{5}, Y in{1}\/{4} ? 
%

% X #= c #<=> B megvalstsa, 1 =< X,c =< 6 esetre.
beq(X, C, B) :- 
  X in 1..6, call(I #= X+6-C), 
  element(I, [0,0,0,0,0,1,0,0,0,0,0], B).

% | ?- beq(X, 5, B).
% 
% X in 1..6,
% B in 0..1 ? 
% 
% yes
% | ?- beq(X, 5, B), X #< 4.
% 
% B = 0,
% X in 1..3 ? 
% 
% yes
% | ?- beq(X, 5, B), B = 0.
% 
% B = 0,
% X in(1..4)\/{6} ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

'abs(x-y)>1'(X,Y) :-
  relation(X, [0-(2..5), 1-(3..5), 2-{0,4,5},
               3-{0,1,5}, 4-(0..2), 5-(0..3)], Y).

% | ?- 'abs(x-y)>1'(X,Y).
% 
% X in 0..5,
% Y in 0..5 ? 
% 
% yes
% | ?- 'abs(x-y)>1'(X,Y), domain([X,Y], 2, 4).
% 
% X in{2}\/{4},
% Y in{2}\/{4} ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

% | ?- domain([X,Y,Z], 1, 2), all_distinct([X,Y,Z]).
% no

%---------------------------------------------------------------------------

% | ?- domain([X,P], 2, 4000), P*X#=7373, minimize(labeling([], [P,X]), P).
% 
% P = 73,
% X = 101 ? ;
% 
% no

%---------------------------------------------------------------------------

schedule(Ss, End) :-
        length(Ss, 7),
        Ds = [16, 6,13, 7, 5,18, 4],
        Rs = [ 2, 9, 3, 7,10, 1,11],
        domain(Ss, 1, 30),
        domain([End], 1, 50),
        after(Ss, Ds, End),
        cumulative(Ss, Ds, Rs, 13),
        labeling([ff,minimize(End)], [End|Ss]).

after([], [], _).
after([S|Ss], [D|Ds], E) :- E #>= S+D, after(Ss, Ds, E).

%% End of file

% | ?- schedule(Ss, End).
% 
% Ss = [1,17,10,10,5,5,1],
% End = 23 ? 

%---------------------------------------------------------------------------

% :- module(tsp, [tsp/3]).
% :- use_module(library(clpfd)).
:- use_module(library(lists), [append/3]).

tsp(Lab, Successor, Cost) :-
        Successor = [X1,X2,X3,X4,X5,X6,X7],
        Costs = [C1,C2,C3,C4,C5,C6,C7],
        element(X1, [0,205,677,581,461,878,345], C1),
        element(X2, [205,0,882,427,390,1105,540], C2),
        element(X3, [677,882,0,619,316,201,470], C3),
        element(X4, [581,427,619,0,412,592,570], C4),
        element(X5, [461,390,316,412,0,517,190], C5),
        element(X6, [878,1105,201,592,517,0,691], C6),
        element(X7, [345,540,470,570,190,691,0], C7),
        sum(Costs, #=, Cost),
        Predecessor = [Y1,Y2,Y3,Y4,Y5,Y6,Y7],
        Costs2 = [D1,D2,D3,D4,D5,D6,D7],
        element(Y1, [0,205,677,581,461,878,345], D1),
        element(Y2, [205,0,882,427,390,1105,540], D2),
        element(Y3, [677,882,0,619,316,201,470], D3),
        element(Y4, [581,427,619,0,412,592,570], D4),
        element(Y5, [461,390,316,412,0,517,190], D5),
        element(Y6, [878,1105,201,592,517,0,691], D6),
        element(Y7, [345,540,470,570,190,691,0], D7),
        sum(Costs2, #=, Cost),
        circuit(Successor, Predecessor),
        append(Successor, Predecessor, All),
        labeling([minimize(Cost)|Lab], All).


% | ?- tsp([ff], Succs, Cost).
% 
%         Cost = 2276, Succs = [2,4,5,6,7,3,1] ? 

%---------------------------------------------------------------------------


'x+y=t'(X,Y,T) +:
        X in min(T) - max(Y)..max(T) - min(Y),
        Y in min(T) - max(X)..max(T) - min(X),
        T in min(X) + min(Y)..max(X) + max(Y) .

% | ?- 'x+y=t'(X,Y,T).
% 
% X in inf..sup,
% Y in inf..sup,
% T in inf..sup ? 
% 
% yes
% | ?- 'x+y=t'(X,Y,T), domain([X,Y,T], 3, 8).
% 
% X in 3..5,
% Y in 3..5,
% T in 6..8 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

't+u=c'(T,U,C) +:                 % Tartomny-konzisztens!
        T in C - dom(U),       U in C - dom(T).

% | ?- T in {1,3,7}, 't+u=c'(T,U,C).
% 
% T in{1}\/{3}\/{7},
% U in inf..sup,
% C in inf..sup ? 
% 
% yes
% | ?- T in {1,3,7}, T+U #= 10.
% 
% T in{1}\/{3}\/{7},
% U in{3}\/{7}\/{9} ? 
% 
% yes
% | ?- 


't+u>=c'(T,U,C) +:
        T in C - max(U)..sup,  U in C - max(T)..sup.

% | ?- T in {1,3,7}, 't+u>=c'(T,U,100).
% 
% T in{1}\/{3}\/{7},
% U in 93..sup ? 
% 
% yes
% | ?- 

'ax+c=t'(A,X,C,Y) +:
        X in (min(Y) - C) /> A .. (max(Y) - C) /< A,
        Y in  min(X)*A + C     .. max(X)*A + C.


% | ?- X in 1..5, 'ax+c=t'(2,X,3,Y).
% 
% X in 1..5,
% Y in 5..13 ? 
% 
% yes
% | ?- Y in 1..5, 'ax+c=t'(2,X,3,Y).
% 
% Y in 1..5,
% X in-1..1 ? 
% 
% yes
% | ?- Y in 6..10, 'ax+c=t'(2,X,3,Y).
% 
% Y in 7..9,
% X in 2..3 ? 
% 
% yes
% | ?- 

'p\\/q'(P, Q, B) +:
        B in (min(P)+min(Q))/>2 .. (max(P)+max(Q))/>2,
        P in min(B)*2 - max(Q) - 1  .. max(B)*2 - min(Q),
        Q in min(B)*2 - max(P) - 1  .. max(B)*2 - min(P).

% | ?- domain([P,Q,B], 0, 1), 'p\\/q'(P, Q, B), P=1.
% 
% B = 1,
% P = 1,
% Q in 0..1 ? 
% 
% yes
% | ?- 

'abs(x-y)>c'(X, Y, C) +:
        X in (inf .. max(Y)-C-1) \/ (min(Y)+C+1 .. sup),
        Y in (inf .. max(X)-C-1) \/ (min(X)+C+1 .. sup).


% | ?- 'abs(x-y)>c'(X,Y,3), X in 3..5.
% 
% X in 3..5,
% Y in(inf..1)\/(7..sup) ? 
% 
% yes
% | ?-

no_threat(no_threat3, X, Y, I) :-
	no_threat_3(X, Y, I).

no_threat_3(X, Y, I) +:
        X in \({Y} \/ {Y+I} \/ {Y-I}),
        Y in \({X} \/ {X+I} \/ {X-I}).


% | ?- time(queens(80,W,no_threat1), T).
% 
% T = 2000,
% W = [1,3,5,44,42,4,50,7,68,57|...] ? 
% 
% yes
% | ?- time(queens(80,W,no_threat3), T).
% 
% T = 280,
% W = [1,3,5,44,42,4,50,7,68,57|...] ? 
% 
% yes
% | ?- 

% Hibs constraint, srti az alapszablyt!
'x=<y=<z'(X, Y, Z) +:
        Y in min(X)..max(Z),
        Z in min(Y).. sup,
        X in inf..max(Y).

% | ?- 'x=<y=<z'(15, 5, Z).
%
% Z in 5..sup ? 
% 
% yes
% | ?-

no_threat(no_threat4, X, Y, I) :-
	no_threat_4(X, Y, I).

% Maximlisan szkt, de nagyon hatkonytalan!
no_threat_4(X, Y, I) +:
        X in unionof(B,dom(Y),\({B} \/ {B+I} \/ {B-I})),
        Y in unionof(B,dom(X),\({B} \/ {B+I} \/ {B-I})).


% | ?- time(queens(80,W,no_threat4), T).
% 
% T = 50600,
% W = [1,3,5,44,42,4,50,7,70,57|...] ? 
% 
% yes
% | ?- 


%---------------------------------------------------------------------------

'x=<y=<z*'(X, Y, Z) +:       % Ez mr helyes!
        Y in min(X)..max(Z),
        Z in ((inf..max(Y)) /\ dom(X)) ? (min(Y).. sup),
        X in ((min(Y)..sup) /\ dom(Z)) ? (inf..max(Y)).

% | ?- 'x=<y=<z*'(15, 5, Z).
% 
% no
% | ?- 

%---------------------------------------------------------------------------

no_threat(no_threat5, X, Y, I) :-
	no_threat_5(X, Y, I).

% Maximlisan szkt, nem nagyon hatkonytalan
no_threat_5(X, Y, I) +:
    X in (4..card(Y)) ? (inf..sup) \/
          unionof(B,dom(Y),\({B} \/ {B+I} \/ {B-I})),
    Y in (4..card(X)) ? (inf..sup) \/
          unionof(B,dom(X),\({B} \/ {B+I} \/ {B-I})).


% | ?- time(queens(80,W,no_threat5), T).
% 
% T = 1900,
% W = [1,3,5,44,42,4,50,7,70,57|...] ? 
% 
% yes
% | ?-

%---------------------------------------------------------------------------

p(X, Y) :- relation(X, [1-{1},2-{1,2},3-{1,2,3}], Y).

q(X, Y) +:
  X in unionof(B,dom(Y),switch(B,[1-{1,2,3},2-{2,3},3-{3}])),
  Y in unionof(B,dom(X),switch(B,[1-{1},2-{1,2},3-{1,2,3}])).

% | ?- X in 1..2, Y in 2..3, p(X, Y).
% 
% X = 2,
% Y = 2 ? ;
% 
% no
% | ?- X in 1..2, Y in 2..3, q(X, Y).
% 
% X = 2,
% Y = 2 ? ;
% 
% no
% | ?- 

%---------------------------------------------------------------------------

'x\\=y'(X,Y) +:        % 1. a constraintet szkit indexiklisok
        X in \{Y},
        Y in \{X}.
'x\\=y'(X,Y) -:        % 2. a negltjt szkit indexiklisok
        X in dom(Y),
        Y in dom(X).
'x\\=y'(X,Y) +?        % 3. a levezethetsget krdez
        X in \dom(Y).  % indexiklis
% 'x\\=y'(X,Y) -?      % 4. a neglt levezethetsgt krdez
%       X in {Y}.      % indexiklis (elhagyhat)


% | ?- X = 10, Y in 1..20, 'x\\=y'(X, Y).
% 
% X = 10,
% Y in(1..9)\/(11..20) ? 
% 
% yes
% | ?- X = 10, 'x\\=y'(X, Y) #<=> B, B = 1.
% 
% B = 1,
% X = 10,
% Y in(inf..9)\/(11..sup) ? 
% 
% yes
% | ?- X in {1,3,7}, 'x\\=y'(X, Y) #<=> B, B = 0.
% 
% B = 0,
% X in{1}\/{3}\/{7},
% Y in{1}\/{3}\/{7} ? 
% 
% yes
% | ?- X in {1,3,7}, 'x\\=y'(X, Y) #<=> B, Y in {2,4,6}.
% 
% B = 1,
% X in{1}\/{3}\/{7},
% Y in{2}\/{4}\/{6} ? 
% 
% yes
% | ?- X in {1,3,7}, 'x\\=y'(X, Y) #<=> B, Y=3, X=3.
% 
% B = 0,
% X = 3,
% Y = 3 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

'x=<y'(X,Y) +:
        X in inf..max(Y),
        Y in min(X)..sup.
'x=<y'(X,Y) -:
        X in (min(Y)+1)..sup,
        Y in inf..(max(X)-1).
'x=<y'(X,Y) +?
        X in inf..min(Y).
'x=<y'(X,Y) -?
        X in (max(Y)+1)..sup.


% | ?- X in {1,3,7}, 'x=<y'(X, Y).
% 
% X in{1}\/{3}\/{7},
% Y in 1..sup ? 
% 
% yes
% | ?- X in {1,3,7}, 'x=<y'(X, Y) #<=> B, B = 1.
% 
% B = 1,
% X in{1}\/{3}\/{7},
% Y in 1..sup ? 
% 
% yes
% | ?- X in {1,3,7}, 'x=<y'(X, Y) #<=> B, B = 0.
% 
% B = 0,
% X in{1}\/{3}\/{7},
% Y in inf..6 ? 
% 
% yes
% | ?- X in {1,3,7}, 'x=<y'(X, Y) #<=> B, Y #> 6.
% 
% B = 1,
% X in{1}\/{3}\/{7},
% Y in 7..sup ? 
% 
% yes
% | ?- X in {1,3,7}, 'x=<y'(X, Y) #<=> B, Y #< 2.
% 
% X in{1}\/{3}\/{7},
% Y in inf..1,
% B in 0..1 ? 
% 
% yes
% | ?- X in {1,3,7}, 'x=<y'(X, Y) #<=> B, Y #< 1.
% 
% B = 0,
% X in{1}\/{3}\/{7},
% Y in inf..0 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------


exactly(I, Xs, N) :-
    dom_susps(Xs, Susp), fd_global(exactly(I,Xs,N), Xs/N, Susp).

dom_susps([], []).
dom_susps([X|Xs], [dom(X)|Susp]) :- dom_susps(Xs, Susp).

:- multifile clpfd:dispatch_global/4.
:- discontiguous clpfd:dispatch_global/4.
clpfd:dispatch_global(exactly(I,_,_), Xs0/N0, Xs/N, Actions) :-
    ex_filter(Xs0, Xs, N0, N, I), length(Xs, M),
    (   N=:=0 -> Actions = [exit|Ps], ex_neq(Xs, I, Ps)
    ;   N=:=M -> Actions = [exit|Ps], ex_eq(Xs, I, Ps)
    ;   N>0, N<M -> Actions = []
    ;   Actions = [fail]
    ).

ex_filter([], [], N, N, _).
ex_filter([X|Xs], Ys, L, N, I) :-
    X==I, !, M is L-1, ex_filter(Xs, Ys, M, N, I).
ex_filter([X|Xs], Ys0, L, N, I) :-
    fd_set(X, Set), fdset_member(I, Set), !,
    Ys0 = [X|Ys], ex_filter(Xs, Ys, L, N, I).
ex_filter([_|Xs], Ys, L, N, I) :- ex_filter(Xs, Ys, L, N, I).

ex_neq(Xs, I, Ps) :- fdset_singleton(Set0, I),
    fdset_complement(Set0, Set), eq_all(Xs, Set, Ps).

ex_eq(Xs, I, Ps) :- fdset_singleton(Set, I), eq_all(Xs, Set, Ps).

eq_all([], _, []).
eq_all([X|Xs], Set, [X in_set Set|Ps]) :- eq_all(Xs, Set, Ps).


% | ?- exactly(5,[A,B,C],1), A=5.
% 
% A = 5,
% B in(inf..4)\/(6..sup),
% C in(inf..4)\/(6..sup) ? 
% 
% yes
% | ?- exactly(5,[A,B,C],1), A in 1..2, B in 3..4.
% 
% C = 5,
% A in 1..2,
% B in 3..4 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

exactly2(I, Xs, N) :-
        dom_susps(Xs, Susp),  
        length(Xs, Len), N in 0..Len,
        fd_global(exactly2(I,Xs,N), Xs/0, [minmax(N)|Susp]).

% :- multifile clpfd:dispatch_global/4.
clpfd:dispatch_global(exactly2(I,_,N), Xs0/D0, Xs/D, Actions) :-
        exactly_solver(I, Xs0, Xs, N, D0, D, Actions).

exactly_solver(I, Xs0, Xs, N, D0, D, Actions) :-
        ex_filter(Xs0, Xs, D0, D, I),
        length(Xs, M), NewMin is -D, NewMax is NewMin+M,
        fd_min(N, MinN), fd_max(N, MaxN), 
        (   MaxN=:=NewMin ->
            Actions = [exit,N=MaxN|Ps], ex_neq(Xs, I, Ps)
        ;   MinN=:=NewMax ->
            Actions = [exit,N=MinN|Ps], ex_eq(Xs, I, Ps)
        ;   Actions = [N in NewMin..NewMax]
        ).

% | ?- _L=[A,B,C,D], domain(_L, 1, 4), 
%      A #< B, B #=< C, C#<D, exactly2(4, _L, N). 
% 
% A in 1..2,
% B in 2..3,
% C in 2..3,
% D in 3..4,
% N in 0..1 ? 
% 
% yes
% | ?- 

%---------------------------------------------------------------------------

% | ?- L = [N,1], N in {0,2}, exactly2(0, L, N).
% 
% L = [0,1],
% N = 0 ? ;
% 
% no
% | ?- spy exactly_solver, L = [N,1], N in {0,2}, exactly2(0, L, N).
% {The debugger will first zip -- showing spypoints (zip)}
% {Plain spypoint for user:exactly_solver/7 added, BID=1}
%  +      1      1 Call: exactly_solver(0,[_1254,1],_1313,_1254,0,_1314,_1300) ? @
% | :- exactly_solver(_,_,_,N,_,_,_)^(fd_set(N, Set), write(Set), nl).
% [[0|0],[2|2]]
%  +      1      1 Call: exactly_solver(0,[_1254,1],_1313,_1254,0,_1314,_1300) ? s
%  +      1      1 Exit: exactly_solver(0,[_1254,1],[_1254],_1254,0,0,[_1254 in 0..1]) ? 
% 
% L = [0,1],
% N = 0 ? ;
% 
% no
% {zip}
% | ?- 

%---------------------------------------------------------------------------

exactly3(I, Xs, N0) :-
        dom_susps(Xs, Susp),  
        length(Xs, Len), N0 in 0..Len, N #= N0,
        fd_global(exactly2(I,Xs,N), Xs/0, [minmax(N)|Susp]).

% | ?- L = [N,1], N in {0,2}, exactly3(0, L, N).
%
% no
% | ?-
